""" ssm.cli.options (boilerplate for click)

    Common CLI arguments for reuse
"""

from functools import partial

import click

key = click.option("--key", help="key to use", default="", required=False)

top = click.option(
    "--top", help="triage and only describe top N results", required=False
)
existing_file = click.option("--file", type=click.Path(exists=True))
filter_service_partial = partial(
    click.option, "--service", help="filter for service", required=False
)
filter_service = filter_service_partial(default=".*")

output_format_partial = partial(
    click.option,
    "--format",
    show_default=True,
    type=click.Choice(["json", "yaml", "yml", "env", "stdout", "tree"]),
    help="output format",
)
output_format = format = output_format_partial(required=True)
output_format_yaml_default = output_format_partial(required=False, default="yaml")
output_format_json_default = output_format_partial(required=False, default="json")
output_format_stdout_default = output_format_partial(required=False, default="stdout")
output_format_tree_default = output_format_partial(required=False, default="tree")


optional_user = user = click.option(
    "--user",
    help="username (default will attempt auto-detect)",
    required=False,
    default="",
)
optional_users = users = click.option(
    "--users", help="user list (comma-separatted)", required=False, default=""
)

raw = click.option(
    "--raw", "-r", help="unquotes text return values (like jq -r)", default=False
)

force_or_dry = force = dry_run = click.option(
    "--force/--dry-run", "-f/-d", help="perform actions", default=False, envvar="FORCE"
)

pause_or_nowait = no_wait = pause = click.option(
    "--pause/--no-wait", help="pause before starting", default=True
)

optional_prefix = click.option(
    "--prefix", help="src prefix to operate under", required=False, default=""
)

cascade_partial = partial(
    click.option,
    "--cascade/--no-cascade",
    "-c/",
    required=False,
    help="cascading search",
)
cascade = cascade_partial(default=True)
no_cascade = cascade_partial(default=False)

flat_output = click.option(
    "--flat-output",
    is_flag=True,
    show_default=True,
    default=False,
    help="flattens output of `/deeply/nested/paths` as simply `path`",
)
dirs_only = click.option(
    "--dirs-only",
    is_flag=True,
    show_default=True,
    default=False,
    help="returns directories only",
)
caller_context = click.option(
    "--caller-context",
    required=False,
    default=False,
    show_default=True,
    is_flag=True,
    help="display details about caller-context",
)

profile_partial = partial(
    click.option,
    "--profile",
    envvar="AWS_PROFILE",
    help="AWS profile to use",
)
src_profile_partial = partial(
    click.option,
    "--src-profile",
    help="Source Environment name",
    show_default=True,
)
dst_profile_partial = partial(
    click.option,
    "--dst-profile",
    help="Destination Environment name",
    show_default=True,
)

required_profile = profile_partial(
    required=True,
)
profile = default_profile = profile_partial(
    default="default",
    required=False,
)
src_profile = src_profile_default = src_profile_partial(
    default="default", required=False
)
dst_profile = dst_profile_default = dst_profile_partial(
    default="default", required=False
)
src_profile_no_default = src_profile_partial(default=None, required=False)
dst_profile_no_default = dst_profile_partial(default=None, required=False)

src_prefix = click.option(
    "--src-prefix",
    envvar="SRC_PREFIX",
    help="src prefix to operate under",
    required=False,
    default="",
)



""" ssm.util
"""

import logging
import functools

import termcolor
import coloredlogs
from rich.text import Text
from rich.tree import Tree

from rich import print as rich_print  # noqa

blue = functools.partial(termcolor.colored, color="blue")
red = functools.partial(termcolor.colored, color="red")
green = functools.partial(termcolor.colored, color="green")
yellow = functools.partial(termcolor.colored, color="yellow")
bold = functools.partial(termcolor.colored, attrs=["bold"])


def rich_walk_dict(dct, tree: Tree, branch_color="[bold magenta]") -> None:
    """recursively build rich.tree.Tree from dict contents"""
    for k, v in dct.items():
        if isinstance(v, (dict,)):
            style = "dim"
            branch = tree.add(
                f"{branch_color}{k}",
                style=style,
                guide_style=style,
            )
            rich_walk_dict(v, branch)
        else:
            tree.add(Text(f"{k}", "green") + Text(": ") + Text(f"{v}"))


def is_string(obj) -> bool:
    """ """
    return isinstance(obj, str)


def flatten_output(result: dict) -> dict:
    """ """
    acc = {}
    for k, v in result.items():
        tmp = k.split("/")
        acc[tmp[-1]] = v
    return acc


def fatal_error(msg):
    """ """
    LOGGER.info("{} {}".format(red("error:"), msg))
    raise SystemExit(1)


def get_logger(name):
    """
    utility function for returning a logger
    with standard formatting patterns, etc
    """
    import os

    if os.path.exists(name):
        name = name.replace(os.path.expanduser("~/"), "")

    class DuplicateFilter(logging.Filter):
        def filter(self, record):
            # add other fields if you need more granular comparison, depends on your app
            current_log = (record.module, record.levelno, record.msg)
            if current_log != getattr(self, "last_log", None):
                self.last_log = current_log
                return True
            return False

    formatter = coloredlogs.ColoredFormatter(
        fmt=" - ".join(
            [
                # "[%(asctime)s]",
                "%(levelname)s\t",
                "%(name)s\t",
                "%(message)s",
            ]
        ),
        # datefmt="%Y-%m-%d %H:%M:%S",
    )
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(formatter)
    logger = logging.getLogger(name)
    if not logger.handlers:
        # prevents duplicate registration
        logger.addHandler(log_handler)
    logger.addFilter(DuplicateFilter())  # add the filter to it
    # FIXME: get this from some kind of global config
    logger.setLevel("DEBUG")
    return logger


LOGGER = get_logger(__name__)



""" ssm.api

  See the docs here:
    https://github.com/Robot-Wranglers/aws-ssm-tool
"""

import collections

import botocore

from ssm import util
from ssm.api.environment import Environment

LOGGER = util.get_logger(__name__)


def _get_env(profile=None, env=None, **kwargs):
    """gets environment from environment or named profile"""
    assert profile or env, str(kwargs)
    env = Environment.from_profile(profile) if profile else env
    assert env
    return env


def _get_client(**kwargs):
    """gets handle for the secrets-manager"""
    env = _get_env(**kwargs)
    env.logger.info("getting client")
    sman = env.secrets
    return sman


def read(secret_name, **kwargs):
    """
    reads a secret
    """
    assert secret_name, f"cannot read secret_name `{secret_name}`"
    secrets = _get_client(**kwargs)
    try:
        return secrets[secret_name]
    except KeyError as exc:
        LOGGER.error(f"KeyError: {exc}")
        raise SystemExit(1)


def stat(path_prefix="/", caller_context: bool = True, **kwargs):
    """
    reports status, including account details and metadata summary for SSM parameters.
    """
    env = _get_env(**kwargs)
    caller_id = env.caller_id
    result = collections.OrderedDict()
    if caller_context:
        LOGGER.info("lookup caller-context")
        result.update(
            context=dict(
                user=dict(
                    id=caller_id.get("UserId"),
                    arn=caller_id.get("Arn"),
                ),
                account=dict(
                    profile_name=env.profile_name,
                    id=env.account_id,
                    alias=env.account_alias,
                    region_name=env.region_name,
                ),
            )
        )
    result.update(
        parameters=dict(
            root=path_prefix,
            key_count=len(env.secrets.under(path_prefix)),
            children=env.secrets.children(path_prefix),
        ),
    )
    return result


import typing


def list_dirs(path_prefix, **kwargs) -> typing.List:
    """
    lists subpaths (i.e. directories) under the given path.
    no leaf nodes are returned
    """
    kwargs.update(dirs_only=True)
    return list(path_prefix, **kwargs)


def list(path_prefix, dirs_only: bool = False, **kwargs) -> typing.List:
    """
    Lists parameters with prefixes below the given path.

    This is recursive by default and only returns leafs:
    use `list-dirs` or pass `--dirs` to get directories only.
    """
    # WARNING: do not use `list()` builtin here..
    secrets = _get_client(**kwargs)
    if dirs_only:
        return secrets.children(path_prefix)
    result = secrets.under(path_prefix).keys()
    result = [x for x in result]
    return result


def get_many(namespace, flat_output: bool = False, **kwargs):
    """
    gets many secrets from specified hierarchy/namespace
    """
    secrets = _get_client(**kwargs)
    result = secrets.under(namespace)
    if flat_output:
        result = util.flatten_output(result)
    return result


def delete(secret_name, no_backup=False, **kwargs):
    """deletes secret (keeping a local-backup is default)"""

    def get_backup_file(prefix):
        return ".tmp.{}".format(prefix.replace("/", "_"))

    secrets = _get_client(**kwargs)
    try:
        parameter = read(secret_name=secret_name, **kwargs)
    except (botocore.exceptions.ClientError,) as exc:
        LOGGER.warning(f"error reading secret @ `{secret_name}` (is this a path?)")
        parameter = None
    if parameter is not None:
        if not no_backup:
            backup = get_backup_file(secret_name)
            LOGGER.debug(f"backup to: {backup}")
            with open(backup, "w") as fhandle:
                fhandle.write(parameter)
        del secrets[secret_name]
        return parameter
    else:
        return False


def move(
    src_name,
    dst_name,
    src_profile: str = "default",
    dst_profile: str = "default",
    **kwargs,
):
    """
    moves a secret from src to dest
    """
    result = copy(
        src_name, dst_name, src_profile=src_profile, dst_profile=dst_profile, **kwargs
    )
    if result:
        src_profile = Environment.from_profile(src_profile)
        del src_profile.secrets[src_name]
    return result


def move_many(
    src_name,
    dst_name,
    src_profile: str = "default",
    dst_profile: str = "default",
    **kwargs,
):
    """
    moves a whole path of secrets to a new location
    """
    dst_name = dst_name[:-1] if dst_name.endswith("/") else dst_name
    src_man = _get_client(profile=src_profile)
    names = [n for n in src_man.under(src_name).keys()]
    LOGGER.debug(f"moving {len(names)} parameters under `{src_name}`")
    results = []
    for name in names:
        new_name = "/".join([dst_name, name[len(src_name) :]])
        LOGGER.debug(f"  {name} -> {new_name}")
        results.append(
            move(
                name,
                new_name,
                src_profile=src_profile,
                dst_profile=dst_profile,
                **kwargs,
            )
        )
    return results


def copy(
    src_name,
    dst_name,
    src_profile: str = "default",
    dst_profile: str = "default",
    **kwargs,
):
    """
    copies a secret from given source to destination
    """
    # NB: mind the signature, this code is reused by `move`
    src_man = _get_client(profile=src_profile)
    dst_man = _get_client(profile=dst_profile)
    dst_name = dst_name or src_name
    try:
        value = src_man[src_name]
    except (botocore.exceptions.ClientError, KeyError) as exc:
        LOGGER.error(f"Cant retrieve `{src_name}` using profile `{src_man.env.name}`!")
        LOGGER.warning(f"Hint: Use `ssm copy-many ...` if `{src_name}` is a hierarchy")
        return False
    else:
        dst_man[dst_name] = value
        return True


def update(secret_name, value, file=None, **kwargs):
    """
    updates secret in given location with new value
    """
    if all([not value, not file]):
        err = (
            "when `value` is not given as second arg, "
            "then `input_file` must be provided"
        )
        LOGGER.critical(err)
        raise RuntimeError(err)
    value = value or open(file).read()
    secrets = _get_client(**kwargs)
    secrets[secret_name] = value
    return True


def delete_path(path_prefix, **kwargs):
    """ """
    raise NotImplementedError()


def put_many(secret_name, input_file=None, **kwargs):  # noqa
    """
    put many secrets
    """
    raise NotImplementedError()



""" ssm.cli.wrapper (click boilerplate)
    Reusable wrapper for building CLIs
"""

import json
import functools

import yaml
import click

from ssm import abcs, util

LOGGER = util.get_logger(__file__)


class ApiWrapper(abcs.Loggable):
    """
    a wrapper that turns a API function into a click CLI subcommand
    """

    BASE_OPTIONS = [
        # every command gets the --debug option
        click.option(
            "--debug",
            default=False,
            is_flag=True,
            help="Enables verbose mode.",
        ),
    ]

    def __init__(
        self,
        command_name=None,
        publishers=[],
        subcommand_name=None,
        fxn=None,
        extra_options=None,
        aliases=[],
        help=None,
        entry=None,
    ):
        self.entry = entry
        self.aliases = aliases
        self.is_subcommand = isinstance(self.entry, (click.core.Group,))
        self.is_stand_alone = self.entry is None
        self.subcommand_name = self.name = subcommand_name or fxn.__name__.replace(
            "_", "-"
        )
        self.command_name = command_name
        self.fxn = fxn
        self.fxn.publishers = getattr(self.fxn, "publishers", publishers)
        self.extra_options = extra_options
        default_help = "no docstring"
        self.help = help or getattr(fxn, "__doc__", default_help) or default_help
        self.help = self.help.strip()
        self.proxy = self.get_proxy()
        if not callable(self.proxy):
            err = f"Expected callable for proxy, got '{self.proxy}'"
            raise ValueError(err)
        if not (self.is_subcommand or self.is_stand_alone):
            err = (
                "expected a group or a standalone "
                "command, got {} of type {} for entry"
            )
            err = err.format(self.entry, type(self.entry))
            raise ValueError(err)
        super().__init__()

    def get_proxy(self):
        """ """
        options = self.extra_options
        if self.is_subcommand:
            # otherwise base options would be added twice for stand-alone style CLIs
            options += self.__class__.BASE_OPTIONS

        @functools.wraps(self.fxn)
        def proxy(*args, **kwargs):
            """ """
            args = [x for x in args if not isinstance(x, (click.core.Context,))]
            # LOGGER.info(f"proxying args={args}, kwargs={kwargs}")
            format = kwargs.get("format", "stdout")
            result = self.fxn(*args, **kwargs)
            if format in ["yaml", "yml"]:
                print(yaml.dump(result))
            elif format in ["json"]:
                print(json.dumps(result))
            elif format in ["python"]:
                print(result)
            elif format in ["env"]:
                assert isinstance(result, (dict,)), f"expected dict, got {type(result)}"
                acc = []
                for k, v in result.items():
                    if isinstance(v, (str,)) and " " in v:
                        msg = "user requested format is `env` but value has a space character!"
                        LOGGER.warning(msg)
                    tmp = "=".join([k.split("/")[-1], v])
                    acc.append(tmp)
                print("\n".join(acc))
            elif format in ["stdout", "tree"]:
                if isinstance(result, (type([]),)):
                    # if format=='tree':
                    #     tree={}
                    #     for path_prefix in result:
                    #         for comp in path_prefix.split('/'):
                    #             tree[path_prefix]=
                    # else:
                    print("\n".join([str(x) for x in result]))
                elif isinstance(result, (dict,)):
                    tree = util.Tree(
                        "",
                        guide_style="bold bright_blue",
                    )
                    util.rich_walk_dict(result, tree)
                    util.rich_print(tree)
                else:
                    util.rich_print(result)
            else:
                raise RuntimeError(f"unrecognized output format `{format}`")

        for option in options:
            proxy = option(proxy)

        if self.is_subcommand:
            if self.aliases:
                result = self.entry.command(
                    name=self.name, help=self.help, aliases=self.aliases
                )(proxy)
            else:
                result = self.entry.command(name=self.name, help=self.help)(proxy)
        elif self.entry is None:
            result = click.command()(proxy)
        else:
            err = "unknown entry type '{}' for '{}'".format(
                type(self.entry), self.entry
            )
            LOGGER.critical(err)
            raise RuntimeError(err)
        return result



""" ssm.cli.args (boilerplate for click)

    Common CLI arguments for reuse
"""

import click

secret_name = click.argument("secret_name", nargs=1)
path_prefix = click.argument("path_prefix", nargs=1)
namespace = click.argument("namespace", nargs=1)
path_prefix_default = click.argument("path_prefix", nargs=1, default="/")
